#define PROCMON_EXTERN
#include "procmon_info.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/procfs.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#define PROCMON_TIMEN_TO_DOUBLE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_nsec) / 1000000000.0)
#define PROCMON_TIMEU_TO_DOUBLE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_usec) / 1000000.0)
/* on aix, they stuff the usec values but really use nsec values ??? */
#define PROCMON_TIMEU_TO_DOUBLE_BROKE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_usec) / 1000000000.0)
/* ========================================================================

   Name
   ====
   PROCMON_info_get_print

   Purpose
   =======
   Get and print process info tagged by the value of tag (with a start/stop
   state).
   This is used to manually wrap blocks of code to get statistics about it:
        PROCMON_info_get_print( "foo", 0 );
        foo();
        PROCMON_info_get_print( "foo", 1 );

   Arguments
   =========
   tag           Intent: in
                 C type: const char *
                 The name associated with the section of code

   state         Intent: in
                 C type: int
                 The state:
                    0: start
                    1: stop

   Return Values
   =============
   none

   Program Flow
   ============
   1) if using PROCMON_rld
   1.1) turn off sig handler, call rld_handler, turn on sig handler
   2) if pure manual
   2.1) initialize if necessary
   2.2) get and print info

   ======================================================================== */
void PROCMON_util_tagname( char *tag_new, const char *tag, int state );
void PROCMON_util_tagname( char *tag_new, const char *tag, int state )
{
  char *char_p; /* char pointer */
  memset( tag_new, 0, PROCMON_TAG_LENGTH );
  if( state >= 0 )
    {
      sprintf( tag_new, "%d:", state );
    }
  strncat( tag_new, tag, PROCMON_TAG_LENGTH - strlen(tag_new) - 1 );
  char_p = tag_new;
  while( (char_p = strchr( char_p, ' ' )) != NULL )
    {
      *char_p = '_';
    }
}
void PROCMON_info_get_print(
                            const char *tag,
                            int state
                            )
{
  static int initialized = 0; /* if this has been initialized */
  static PROCMON_INFO_struct pinfo;
  static PROCMON_INFO_SETTINGS_struct psettings;
  int procmon_error; /* error return */
  static FILE *output_fd = NULL; /* output file */
  if( state != 0 )
    {
      state = 1;
    }
  /* ------------------ */
  /* PROCMON_rld active */
  /* ------------------ */
  if( procmon_s.initialized )
    {
      /* ------------------------- */
      /* temp turn off sig handler */
      /* ------------------------- */
      if( sigaction(procmon_s.which_sigaction,
                    &(procmon_s.action_ignore), NULL ) != 0 )
        abort();
      /* --------------------------------------------------------- */
      /* set tag and call handler (which will get/print proc info) */
      /* --------------------------------------------------------- */
      PROCMON_util_tagname( procmon_s.psettings.tag, tag, state );
      PROCMON_rld_handler( 0, NULL, NULL );
      PROCMON_util_tagname( procmon_s.psettings.tag, PROCMON_TAG_DEFAULT, -1 );
      /* ------------------------ */
      /* turn back on sig handler */
      /* ------------------------ */
      if (sigaction(procmon_s.which_sigaction,
                    &(procmon_s.action), NULL ) != 0 )
        abort();
    }
  /* ------------------------ */
  /* DONE: PROCMON_rld active */
  /* ------------------------ */
  /* --------------------------- */
  /* pure manual instrumentation */
  /* --------------------------- */
  else
    {
      /* --------------------------------------------------- */
      /* fork workaround - re-init to get new file if needed */
      /* --------------------------------------------------- */
      if( initialized && getpid() != psettings.pid )
        {
          initialized = 0;
        }
      /* ----------------------- */
      /* initialize if necessary */
      /* ----------------------- */
      if( ! initialized )
        {
          PROCMON_init_file( &pinfo, &psettings, &output_fd );
          initialized = 1;
        }
      /* ----------------------------- */
      /* DONE: initialize if necessary */
      /* ----------------------------- */
      /* -------------------------------------------- */
      /* get a valid info then print it (if possible) */
      /* -------------------------------------------- */
      while( (procmon_error = PROCMON_info_get( psettings.pid, &pinfo )) ==
             PROCMON_WARNING ) {};
      if( procmon_error == 0 )
        {
          PROCMON_util_tagname( psettings.tag, tag, state );
          PROCMON_info_print( &pinfo, &psettings, output_fd );
          PROCMON_util_tagname( psettings.tag, PROCMON_TAG_DEFAULT, -1 );
        }
    }
  /* --------------------------------- */
  /* DONE: pure manual instrumentation */
  /* --------------------------------- */
}
void procmon_info_get_print(
                            const char *tag,
                            int *state
                            )
{
  PROCMON_info_get_print( tag, *state );
}
void procmon_info_get_print_(
                             const char *tag,
                             int *state
                             )
{
  PROCMON_info_get_print( tag, *state );
}
void procmon_info_get_print__(
                              const char *tag,
                              int *state
                              )
{
  PROCMON_info_get_print( tag, *state );
}
void PROCMON_INFO_GET_PRINT(
                            const char *tag,
                            int *state
                            )
{
  PROCMON_info_get_print( tag, *state );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_settings_get

   Purpose
   =======
   Get the settings for procmon.  This will process command line
   args or environment variable settings.

   Arguments
   =========
   argc       Intent: in
              C type: int
              Argument count (from main()).
              If 0, will just process environment vars.
              
   argv       Intent: in
              C type: const char**
              Argument list (from main()).
              If NULL, will just process environment vars.

   psettings  Intent: in
              C type: PROCMON_INFO_SETTINGS_struct*
              Struct to be filled.

   ierr       Intent: out
              C type: (na) error return value
              error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) get env var settings.
   2) override with any command line settings

   ======================================================================== */
int PROCMON_info_settings_get(
                              int argc,
                              const char **argv,
                              PROCMON_INFO_SETTINGS_struct *psettings
                              )
{
  int
    i, /* loop var */
    ierr=0; /* error return value */
  char
    *empty_string = "", /* empty string */
    *getenv_char_p, /* pointer to env var */
    hostname[PROCMON_MAX_LEN]; /* the hostname */
  /* ---- */
  /* init */
  /* ---- */
  /* -------- */
  /* defaults */
  /* -------- */
  psettings->time     = 0;
  psettings->interval = 1;
  psettings->min      = 1;
  sprintf( psettings->dir, "%s", empty_string );
  psettings->pid      = 0; /* mininum value (account for unsigned if needed) */
  psettings->pretty   = 0;
  psettings->prettyw  = 0;
  psettings->signal   = PROCMON_SIGNAL_DEFAULT;
  psettings->uid      = -1; /* less than 0 - 0 is root */
  sprintf( psettings->command, "%s", empty_string );
  sprintf( psettings->tag, "%s", PROCMON_TAG_DEFAULT );
  /* -------------------- */
  /* get env var settings */
  /* -------------------- */
  /* ------------ */
  /* PROCMON_TIME */
  /* ------------ */
  getenv_char_p = getenv( "PROCMON_TIME" );
  if( getenv_char_p != NULL )
    {
      psettings->time = atoi( getenv_char_p );
      if( psettings->time < 0 )
        {
          psettings->time = 0;
        }
    }
  /* ---------------- */
  /* PROCMON_INTERVAL */
  /* ---------------- */
  getenv_char_p = getenv( "PROCMON_INTERVAL" );
  if( getenv_char_p != NULL )
    {
      psettings->interval = atoi( getenv_char_p );
      if( psettings->interval < 0 )
        {
          psettings->interval = 0;
        }
    }
  /* ----------- */
  /* PROCMON_MIN */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_MIN" );
  if( getenv_char_p != NULL )
    {
      psettings->min = atoi( getenv_char_p );
      if( psettings->min <= 0 )
        {
          psettings->min = 1;
        }
    }
  /* ----------- */
  /* PROCMON_DIR */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_DIR" );
  if( getenv_char_p == NULL || strlen( getenv_char_p ) <= 0 )
    {
      getenv_char_p = empty_string;
    }
  assert( strlen(getenv_char_p) < PROCMON_MAX_LEN );
  sprintf( psettings->dir, "%s", getenv_char_p );
  /* --------------- */
  /* PROCMON_COMMAND */
  /* --------------- */
  getenv_char_p = getenv( "PROCMON_COMMAND" );
  if( getenv_char_p == NULL || strlen( getenv_char_p ) <= 0 )
    {
      getenv_char_p = empty_string;
    }
  assert( strlen(getenv_char_p) < PROCMON_MAX_LEN );
  sprintf( psettings->command, "%s", getenv_char_p );
  /* ----------- */
  /* PROCMON_UID */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_UID" );
  if( getenv_char_p != NULL )
    {
      psettings->uid = atoi( getenv_char_p );
      if( psettings->uid < 0 )
        {
          psettings->uid = -1; /* less than 0 - 0 is root */
        }
    }
  /* ----------- */
  /* PROCMON_PID */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_PID" );
  if( getenv_char_p != NULL )
    {
      psettings->pid = atoi( getenv_char_p );
      if( psettings->pid < 0 )
        {
          psettings->pid = 0; /* account for unsigned possibility */
        }
    }
  /* -------------- */
  /* PROCMON_PRETTY */
  /* -------------- */
  getenv_char_p = getenv( "PROCMON_PRETTY" );
  if( getenv_char_p != NULL )
    {
      psettings->pretty = 1;
    }
  /* --------------- */
  /* PROCMON_PRETTYW */
  /* --------------- */
  getenv_char_p = getenv( "PROCMON_PRETTYW" );
  if( getenv_char_p != NULL )
    {
      psettings->prettyw = atoi( getenv_char_p );
      if( psettings->prettyw < 0 )
        {
          psettings->prettyw = 0;
        }
    }
  /* -------------- */
  /* PROCMON_SIGNAL */
  /* -------------- */
#if ! defined( PROCMON_SIGNAL_NO_ENV )
  getenv_char_p = getenv( "PROCMON_SIGNAL" );
  if( getenv_char_p != NULL )
    {
      psettings->signal = atoi( getenv_char_p );
      if( psettings->signal <= 0 )
        {
          psettings->signal = PROCMON_SIGNAL_DEFAULT;
        }
    }
#endif
  /* -------------------------- */
  /* DONE: get env var settings */
  /* -------------------------- */
  /* ------------------------- */
  /* get command line settings */
  /* ------------------------- */
  i = 1;
  while( i < argc && argv != NULL )
    {
      /* ----- */
      /* -time */
      /* ----- */
      if( strcmp( argv[i], "-time" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->time = atoi( argv[i+1] );
          if( psettings->time < 0 )
            {
              psettings->time = 0;
            }
          i = i + 2;
        }
      /* --------- */
      /* -interval */
      /* --------- */
      else if( strcmp( argv[i], "-interval" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->interval = atoi( argv[i+1] );
          if( psettings->interval < 0 )
            {
              psettings->interval = 0;
            }
          i = i + 2;
        }
      /* ---- */
      /* -min */
      /* ---- */
      else if( strcmp( argv[i], "-min" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->min = atoi( argv[i+1] );
          if( psettings->min <= 0 )
            {
              psettings->min = 1;
            }
          i = i + 2;
        }
      /* ---- */
      /* -dir */
      /* ---- */
      else if( strcmp( argv[i], "-dir" ) == 0 )
        {
          assert( i < argc - 1 );
          assert( strlen( argv[i+1] ) < PROCMON_MAX_LEN );
          sprintf( psettings->dir, "%s", argv[i+1] );
          i = i + 2;
        }
      /* -------- */
      /* -command */
      /* -------- */
      else if( strcmp( argv[i], "-command" ) == 0 )
        {
          assert( i < argc - 1 );
          assert( strlen( argv[i+1] ) < PROCMON_MAX_LEN );
          sprintf( psettings->command, "%s", argv[i+1] );
          i = i + 2;
        }
      /* ---- */
      /* -uid */
      /* ---- */
      else if( strcmp( argv[i], "-uid" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->uid = atoi( argv[i+1] );
          if( psettings->uid < 0 )
            {
              psettings->uid = -1; /* less than 0 - 0 is root */
            }
          i = i + 2;
        }
      /* ------- */
      /* -pretty */
      /* ------- */
      else if( strcmp( argv[i], "-pretty" ) == 0 )
        {
          psettings->pretty = 1;
          i = i + 1;
        }
      /* -------- */
      /* -prettyw */
      /* -------- */
      else if( strcmp( argv[i], "-prettyw" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->prettyw = atoi( argv[i+1] );
          if( psettings->prettyw < 0 )
            {
              psettings->prettyw = 0;
            }
          i = i + 2;
        }
      /* ------- */
      /* -signal */
      /* ------- */
      else if( strcmp( argv[i], "-signal" ) == 0 )
        {
#if ! defined( PROCMON_SIGNAL_NO_ENV )
          assert( i < argc - 1 );
          psettings->signal = atoi( argv[i+1] );
          if( psettings->signal <= 0 )
            {
              psettings->signal = PROCMON_SIGNAL_DEFAULT;
            }
#endif
          i = i + 2;
        }
      /* --- */
      /* pid */
      /* --- */
      else
        {
          psettings->pid = atoi( argv[i] );
          if( psettings->pid < 0 )
            {
              psettings->pid = 0; /* account for possible unsigned in future */
            }
          i = i + 1;
        }
    }
  /* ---------------------------------------------------- */
  /* default if no qualifiers set use pid of this process */
  /* ---------------------------------------------------- */
  if( psettings->pid == 0 &&
      psettings->uid < 0 &&
      strlen( psettings->command ) == 0 )
    {
      psettings->pid = getpid();
    }
  /* ------------------------------- */
  /* unset pretty if dir set as well */
  /* ------------------------------- */
  if( strlen( psettings->dir ) > 0 )
    {
      psettings->pretty = 0;
    }
  /* ------------------------------------------------- */
  /* get guess at output file prefix and name          */
  /* (procmon might have pid 0 but gets own file name) */
  /* ------------------------------------------------- */
  psettings->outfile_prefix[0] = '\0';
  psettings->outfile_name[0] = '\0';
  if( strlen( psettings->dir ) > 0 )
    {
      if(gethostname(hostname, PROCMON_MAX_LEN-1) != 0)
        abort();
      sprintf( psettings->outfile_prefix,
               "%s/%s%s%s",
               psettings->dir, "procmon_", hostname, "_" );
      sprintf( psettings->outfile_name,
               "%s%llu%s",
               psettings->outfile_prefix, psettings->pid,".txt" );
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_settings_print

   Purpose
   =======
   Print the data from a PROCMON_INFO_SETTINGS_struct to the specified file.

   Arguments
   =========
   psettings  Intent: in
              C type: const PROCMON_INFO_SETTINGS_struct*
              The struct to print out.
             
   outfile    Intent: in
              C type: FILE*
              The file which to print (eg from fopen).
              If NULL, will print to stdout.
             
   ierr       Intent: out
              C type: (na) error return value
              error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) print settings info.

   ======================================================================== */
int PROCMON_info_settings_print(
                                const PROCMON_INFO_SETTINGS_struct *psettings,
                                FILE *outfile
                                )
{
  int ierr=0; /* error return value */
  FILE *of; /* pointer to actual output file */
  of = outfile;
  if( outfile == NULL )
    {
      of = stdout;
    }
  if( psettings != NULL )
    {
      fprintf( of,
               "#PROCMON_SETTINGS:"
               " TIME [%llu]"
               " MIN [%llu]"
               " INTERVAL [%llu]"
               " DIR [%s]"
               "\n",
               psettings->time,
               psettings->min,
               psettings->interval,
               psettings->dir );
      fflush( of );
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_copy

   Purpose
   =======
   Copy data from one PROCMON_INFO_struct to another

   Arguments
   =========
   pinfo         Intent: in
                 C type: const PROCMON_INFO_struct*
                 The original pinfo.

   pinfoc        Intent: out
                 C type: PROCMON_INFO_struct*
                 The copy.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) copy data

   ======================================================================== */
int PROCMON_info_copy(
                      const PROCMON_INFO_struct *pinfo,
                      PROCMON_INFO_struct *pinfoc
                      )
{
  int ierr=0; /* error return value */
  int i; /* loop var */
  pinfoc->num_samples    = pinfo->num_samples;
  pinfoc->num_samples_printed = pinfo->num_samples_printed;
  pinfoc->printed_header = pinfo->printed_header;
  pinfoc->chars_printed  = pinfo->chars_printed;
  pinfoc->wall           = pinfo->wall;
  pinfoc->m_size         = pinfo->m_size;
  pinfoc->m_free         = pinfo->m_free;
  pinfoc->p_pid          = pinfo->p_pid;
  pinfoc->p_ppid         = pinfo->p_ppid;
  pinfoc->p_uid          = pinfo->p_uid;
  strncpy( pinfoc->p_command, pinfo->p_command, PROCMON_MAX_LEN );
  pinfoc->thread         = pinfo->thread;
  pinfoc->p_size_lim     = pinfo->p_size_lim;
  pinfoc->p_rss_lim      = pinfo->p_rss_lim;
  pinfoc->p_data_lim     = pinfo->p_data_lim;
  pinfoc->p_stack_lim    = pinfo->p_stack_lim;
  pinfoc->p_rss          = pinfo->p_rss;
  pinfoc->p_data         = pinfo->p_data;
  pinfoc->p_stack        = pinfo->p_stack;
  pinfoc->p_majflt       = pinfo->p_majflt;
  pinfoc->p_minflt       = pinfo->p_minflt;
  for( i = 0; i < PROCMON_SAMPLE_SAVE; i++ )
    {
      pinfoc->p_size[i] = pinfo->p_size[i];
      pinfoc->p_time[i] = pinfo->p_time[i];
      pinfoc->p_wall[i] = pinfo->p_wall[i];
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_init

   Purpose
   =======
   Initialize a PROCMON_INFO_struct

   Arguments
   =========
   pinfo         Intent: out
                 C type: PROCMON_INFO_struct*
                 The original pinfo.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) copy data

   ======================================================================== */
int PROCMON_info_init(
                      PROCMON_INFO_struct *pinfo
                      )
{
  int ierr = 0; /* error return value */
  int i; /* loop var */
  pinfo->num_samples = 0;
  pinfo->num_samples_printed = 0;
  pinfo->printed_header = 0;
  pinfo->chars_printed = 0;
  pinfo->wall = 0;
  pinfo->m_size = 0;
  pinfo->m_free = 0;
  pinfo->p_pid = 0;
  pinfo->p_ppid = 0;
  pinfo->p_uid = 0;
  memset( pinfo->p_command, '\0', PROCMON_MAX_LEN );
  pinfo->thread = 0;
  pinfo->p_size_lim = 0;
  pinfo->p_rss_lim = 0;
  pinfo->p_data_lim = 0;
  pinfo->p_stack_lim = 0;
  pinfo->p_rss = 0;
  pinfo->p_data = 0;
  pinfo->p_stack = 0;
  pinfo->p_majflt = 0;
  pinfo->p_minflt = 0;
  for( i = 0; i < PROCMON_SAMPLE_SAVE; i++ )
    {
      pinfo->p_size[i] = 0;
      pinfo->p_time[i] = 0;
      pinfo->p_wall[i] = 0;
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_print

   Purpose
   =======
   Print the data from a PROCMON_INFO_struct to the specified file.

   Arguments
   =========
   pinfo         Intent: inout
                 C type: const PROCMON_INFO_struct*
                 The process info struct (filled by PROCMON_info_get).

   psettings     Intent: in
                 C type: const PROCMON_INFO_SETTINGS_struct*
                 Contains settings used to gather/print data.
                 If NULL, defaults will be used.
             
   outfile       Intent: in
                 C type: FILE*
                 The file which to print (eg from fopen).
                 If NULL, will print to stdout.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) print header or print pinfo.

   ======================================================================== */
#define NUM_FONTS 7
int PROCMON_info_print(
                       PROCMON_INFO_struct *pinfo,
                       PROCMON_INFO_SETTINGS_struct *psettings,
                       FILE *outfile
                       )
{
  int i; /* loop var */
  int ierr=0; /* error return value */
  FILE *of; /* pointer to actual output file */
  double pTimeCur[PROCMON_SAMPLE_SAVE];
  static char *tag_default = PROCMON_TAG_DEFAULT;
  char *tag = tag_default;
  int num_cols_h; /* number of columns printed in header */
  int num_cols_d; /* number of columns printed in data */
  char line_head[1024]; /* the line to print */
  char line_data[1024]; /* the line to print */
  char *line_headp; /* pointer to inside line */
  char *line_datap; /* pointer to inside line */
  static char *sbold = "\033[01m";
  static char *sundl = "\033[04m";
  static char *sfred = "\033[31m";
  static char *sfgrn = "\033[32m";
  static char *sfblu = "\033[34m";
  static char *sfmag = "\033[35m";
  static char *sfcyn = "\033[36m";
  static char *snorm = "\033[00m";
  static char *sempt = "";
  char *fonts[NUM_FONTS]; /* a list of fonts to use */
  char *pnorm;
  /* ---- */
  /* init */
  /* ---- */
  of = outfile;
  if( outfile == NULL )
    {
      of = stdout;
    }
  if( psettings && strlen( psettings->tag ) > 0 )
    {
      tag = psettings->tag;
    }
  if( psettings && psettings->pretty != 0 )
    {
      i = 0;
      fonts[i++] = sbold;
      fonts[i++] = sundl;
      fonts[i++] = sfred;
      fonts[i++] = sfgrn;
      fonts[i++] = sfblu;
      fonts[i++] = sfmag;
      fonts[i++] = sfcyn;
      pnorm = snorm;
    }
  else
    {
      i = 0;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      fonts[i++] = sempt;
      pnorm = sempt;
    }
  if( psettings && psettings->pretty != 0 )
    {
      pnorm = snorm;
    }
  else
    {
      pnorm = sempt;
    }
  /* -------------------- */
  /* print general header */
  /* -------------------- */
  if( pinfo->printed_header == 0 )
    {
      fprintf( of,
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n"
               "#PROCMON:key %s%9s%s = %s\n",
               fonts[0], "EpochTime", pnorm, "Time (seconds) Since Epoch",
               fonts[0], "pWallTime", pnorm, "Process Wall Time (seconds)",
               fonts[0], "%pTimeTot", pnorm, "% CPU Usage Total ([User+System]/Wall)",
               fonts[0], "%pTimeCur", pnorm, "% CPU Usage Instantaneous ([User+System]/Wall)",
               fonts[0], "%mMemFree", pnorm, "% Machine Memory Free",
               fonts[0], "pSize",     pnorm, "Process Size (bytes)",
               fonts[0], "pSizeMax",  pnorm, "Maximum Process Size Recorded",
               fonts[0], "pRSS",      pnorm, "Resident Set Size [process size in RAM] (bytes)",
               fonts[0], "%RSS:pSze", pnorm, "% of Process Size   of RSS",
               fonts[0], "%RSS:mMem", pnorm, "% of Machine Memory of RSS",
               fonts[0], "%pLimSize", pnorm, "% of pSize Limit",
               fonts[0], "%pLimRSS",  pnorm, "% of RSS   Limit",
               fonts[0], "%pLimData", pnorm, "% of Data  Limit",
               fonts[0], "%pLimStck", pnorm, "% of Stack Limit",
               fonts[0], "pSizeData", pnorm, "Process Data   Size (bytes)",
               fonts[0], "pSizeStck", pnorm, "Process Stack  Size (bytes)",
               fonts[0], "pMajFlt",   pnorm, "Number of Major Page Faults (disk reads)",
               fonts[0], "pMinFlt",   pnorm, "Number of Minor Page Faults (no disk reads)",
               fonts[0], "MMEMSIZE",  pnorm, "Machine Memory Size (bytes) - Constant",
               fonts[0], "UID",       pnorm, "User ID    - Constant",
               fonts[0], "PID",       pnorm, "Process ID - Constant",
               fonts[0], "TAG",       pnorm, "Sample Tag",
               fonts[0], "COMMAND",   pnorm, "Command    - Constant"
               );
    }
  /* -------------------------- */
  /* DONE: print general header */
  /* -------------------------- */
  /* -------------- */
  /* get some stats */
  /* -------------- */
  if( pinfo->wall > 0 )
    {
      if( pinfo->p_wall[0] == 0 )
        {
          pTimeCur[0] = 0;
        }
      else
        {
          pTimeCur[0] = 100.0*(pinfo->p_time[0]/pinfo->p_wall[0]);
        }
      for( i = 1; i < PROCMON_SAMPLE_SAVE; i++ )
        {
          if( pinfo->p_wall[i-1]-pinfo->p_wall[i] == 0 )
            {
              pTimeCur[i] = 0;
            }
          else
            {
              pTimeCur[i] = 100.0*((pinfo->p_time[i-1]-pinfo->p_time[i])/
                                   (pinfo->p_wall[i-1]-pinfo->p_wall[i]));
            }
        }
      /* -------------- */
      /* "fix" bad data */
      /* -------------- */
      for( i = 0; i < PROCMON_SAMPLE_SAVE; i++ )
        {
          if( pTimeCur[i] < 0.0 )
            {
              pTimeCur[i] = 0;
            }
          if( pTimeCur[i] > 100.0 )
            {
              pTimeCur[i] = 100.0;
            }
        }
    }
  /* -------------------- */
  /* DONE: get some stats */
  /* -------------------- */
  /* -------------------- */
  /* build lines to print */
  /* -------------------- */
  num_cols_d = 0;
  num_cols_h = 0;
  line_datap = line_data;
  line_headp = line_head;
  (pinfo->num_samples_printed)++;
  /* start with PROCMON */
  line_headp += sprintf( line_headp, "%9s ", "#PROCMON " );
  line_datap += sprintf( line_datap, "%9s ", "PROCMON  " );
  /* EpochTime */
  line_headp += sprintf( line_headp, "%s%17s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "EpochTime",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%17f%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->wall,
                         pnorm );
  /* pWallTime */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pWallTime",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_wall[0],
                         pnorm );
  /* %pTimeTot */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "%pTimeTot",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pTimeCur[0],
                         pnorm );
  /* %pTimeCur */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "%pTimeCur",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pTimeCur[1],
                         pnorm );
  /* %mMemFree */
  if( pinfo->m_size > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "%mMemFree",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->m_free/pinfo->m_size),
                             pnorm );
    }
  /* pSize */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pSize",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_size[0],
                         pnorm );
  /* pSizeMax */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pSizeMax",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_size[1],
                         pnorm );
  /* pRSS */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pRSS",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_rss,
                         pnorm );
  /* %RSS:pSze */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "%RSS:pSze",
                         pnorm );
  if( pinfo->p_size[0] > 0 )
    {
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->p_rss/pinfo->p_size[0]),
                             pnorm );
    }
  else
    {
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS], 0.0,
                             pnorm );
    }
  /* %RSS:mMem */
  if( pinfo->m_size > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "%RSS:mMem",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->p_rss/pinfo->m_size),
                             pnorm );
    }
  /* %pLimSize */
  if( pinfo->p_size_lim > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "%pLimSize",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->p_size[0]/pinfo->p_size_lim),
                             pnorm );
    }
  /* %pLimRSS */
  if( pinfo->p_rss_lim > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS], "%pLimRSS",
                             pnorm );
          line_datap += sprintf( line_datap, "%s%9.2e%s ",
                            fonts[(num_cols_d++)%NUM_FONTS],
                            100.0*(pinfo->p_rss/pinfo->p_rss_lim),
                            pnorm );
    }
  /* %pLimData or pSizeData */
  if( pinfo->p_data_lim > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "%pLimData",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->p_data/pinfo->p_data_lim),
                             pnorm );
    }
  else
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "pSizeData",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             pinfo->p_data,
                             pnorm );
    }
  /* %pLimStck or pSizeStck */
  if( pinfo->p_stack_lim > 0 )
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "%pLimStck",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             100.0*(pinfo->p_stack/pinfo->p_stack_lim),
                             pnorm );
    }
  else
    {
      line_headp += sprintf( line_headp, "%s%9s%s ",
                             fonts[(num_cols_h++)%NUM_FONTS],
                             "pSizeStck",
                             pnorm );
      line_datap += sprintf( line_datap, "%s%9.2e%s ",
                             fonts[(num_cols_d++)%NUM_FONTS],
                             pinfo->p_stack,
                             pnorm );
    }
  /* pMajFlt */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pMajFlt",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_majflt,
                         pnorm );
  /* pMinFlt */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "pMinFlt",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_minflt,
                         pnorm );
  /* MMEMSIZE (must be there since this has stopping point in procmon_post.pl) */
  /* Will be -1 if not available */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS],
                         "MMEMSIZE",
                         pnorm );
  line_datap += sprintf( line_datap, "%s%9.2e%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->m_size,
                         pnorm );
  /* UID */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS], "UID",      pnorm );
  line_datap += sprintf( line_datap, "%s%9llu%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_uid,
                         pnorm );
  /* PID */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS], "PID",      pnorm );
  line_datap += sprintf( line_datap, "%s%9llu%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_pid,
                         pnorm );
  /* TAG */
  line_headp += sprintf( line_headp, "%s%9s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS], "TAG",      pnorm );
  line_datap += sprintf( line_datap, "%s%9s%s ",
                         fonts[(num_cols_d++)%NUM_FONTS], tag,
                         pnorm );
  /* COMMAND */
  line_headp += sprintf( line_headp, "%s%s%s ",
                         fonts[(num_cols_h++)%NUM_FONTS], "COMMAND",  pnorm );
  line_datap += sprintf( line_datap, "%s%s%s",
                         fonts[(num_cols_d++)%NUM_FONTS], pinfo->p_command,
                         pnorm );
  /* -------------------------- */
  /* DONE: build lines to print */
  /* -------------------------- */
  /* ----------- */
  /* print lines */
  /* ----------- */
  if( pinfo->printed_header == 0 )
    {
      fprintf( of, "%s\n", line_head );
      pinfo->printed_header = 1;
    }
  if( pinfo->wall > 0 )
    {
      pinfo->chars_printed  = fprintf( of, "%s\n", line_data );
      /* ----------------------------- */
      /* subtract out formatting chars */
      /* ----------------------------- */
      pinfo->chars_printed -= num_cols_d*(strlen(pnorm)*2);
    }
  fflush( of );
  return( ierr );
}
/* ========================================================================

   Name
   ====
   PROCMON_info_get

   Purpose
   =======
   Fill the PROCMON_INFO_struct for the specified pid.

   Arguments
   =========
   pid           Intent: in
                 C type: long long
                 The process id.  If 0, use this processes id.

   pinfo         Intent: inout
                 C type: PROCMON_INFO_struct*
                 The process info struct
                 If NULL, just the header will print out.
                 If there is an error (eg the process went away)
                 this struct will not be modified.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value.
                 If there is an error, pinfo will not be modified.
                 An error might happen if the process cannot be found
                 (ie it terminated) or bad data was detected (eg
                 having process time exceed wall clock time).
                 See the Return Values below.

   Notes
   =====
   This is a collection of machine dependent PROCMON_info_get routines.
   A good deal of this information was obtained by examining the top utility
   from http://sourceforge.net/projects/unixtop/:
     - Concept of creating a machine independent data structure and a
       routine to fill this data.  Machine dependent fill routines are
       called.
     - Information about how to fill the data was mimicked (or coppied) from
       unixtop routines.

   Return Values
   =============
   0 if ok
   PROCMON_ERROR:   Will likely never fill data (eg process gone)
   PROCMON_WARNING: Current sampling is strange (pinfo not filled)

   Program Flow
   ============
   1) fill stuff from a myriad of sources (/proc, ioctl, sysmp, table,
      getrusage, ...)

   ======================================================================== */
/* --- */
/* AIX */
/* --- */
#if defined( _PROCMON_AIX )
#include <procinfo.h>
#include <sys/procfs.h>
/* cannot find include file with this prototype - UGH!!! */
int getprocs( struct procsinfo *, int, struct fdsinfo *, int, pid_t *, int );
int PROCMON_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  long long num_chars; /* number of chars to print in strncpy */
  PROCMON_INFO_struct pinfot; /* if no error, will be coppied to pinfo */
  unsigned long long pagesize;
  struct procsinfo procsinfo_dat;
  pid_t pid_i; /* integer copy of pid - modified */
  struct timespec time_current; /* clock_gettime */
  struct psinfo psinfo_data; /* /proc/<PID>/psinfo */
  struct pstatus pstatus_data; /* /proc/<PID>/status */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  int proc_fid; /* /proc file id */
  /* init */
  PROCMON_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  /* must be root to get machine size and free mem?!?!?! UGH!!! */
  pinfot.m_size = -1;
  pinfot.m_free = -1;
  pagesize = getpagesize();
  /* /proc/<PID>/psinfo */
  sprintf( proc_buf_filename, "/proc/%llu/psinfo", pinfot.p_pid );
  if( (proc_fid = open( proc_buf_filename, O_RDONLY) ) == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone or inaccessible */
      pinfot.p_ppid      = -1;
      pinfot.p_uid       = -1;
      pinfot.p_size[0]   = -1;
      pinfot.p_rss       = -1;
      pinfot.p_time[0]   = -1;
      pinfot.wall        = -1;
      pinfot.p_wall[0]   = -1;
    }
  else
    {
      read( proc_fid, &psinfo_data, sizeof( psinfo_data ) );
      close( proc_fid );
      pinfot.p_ppid      = psinfo_data.pr_ppid;
      pinfot.p_uid       = psinfo_data.pr_uid;
      /* empirical data showed divide by 4...dunno why...getpagesize broke? */
      pinfot.p_size[0]   = (psinfo_data.pr_size * pagesize)/4.0;
      pinfot.p_rss       = (psinfo_data.pr_rssize * pagesize)/4.0;
      pinfot.p_time[0]   = PROCMON_TIMEN_TO_DOUBLE( psinfo_data.pr_time );
      clock_gettime( CLOCK_REALTIME, &time_current );
      pinfot.wall        = PROCMON_TIMEN_TO_DOUBLE( time_current );
      pinfot.p_wall[0]   =
	pinfot.wall - PROCMON_TIMEN_TO_DOUBLE(psinfo_data.pr_start);
      num_chars = strlen( psinfo_data.pr_fname );
      num_chars =
        num_chars < PROCMON_MAX_LEN - 1 ? num_chars : PROCMON_MAX_LEN - 1;
      strncpy( pinfot.p_command, psinfo_data.pr_fname,num_chars );
    }
  /* /proc/<PID>/status */
  sprintf( proc_buf_filename, "/proc/%llu/status", pinfot.p_pid );
  if( (proc_fid = open( proc_buf_filename, O_RDONLY) ) == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone or inaccessible */
      pinfot.p_data      = -1;
      pinfot.p_stack     = -1;
    }
  else
    {
      read( proc_fid, &pstatus_data, sizeof( pstatus_data ) );
      close( proc_fid );
      pinfot.p_data      = pstatus_data.pr_brksize;
      /* always returned 0 - try with getprocs below */
      /* pinfot.p_stack     = pstatus_data.pr_stksize; */
    }
  /* -------- */
  /* getprocs */
  /* -------- */
  memset( &procsinfo_dat, 0, sizeof( procsinfo_dat ) );
  pid_i = pinfot.p_pid;
  if( getprocs(&procsinfo_dat,sizeof(procsinfo_dat),NULL,0,&pid_i,1) != 1 ||
      procsinfo_dat.pi_pid != pinfot.p_pid )
    {
      ierr = PROCMON_ERROR; /* process is gone or inaccessible */
      pinfot.p_stack     = -1;
      pinfot.p_majflt    = -1;
      pinfot.p_minflt    = -1;
      pinfot.p_size_lim  = -1;
      pinfot.p_rss_lim   = -1;
      pinfot.p_data_lim  = -1;
      pinfot.p_stack_lim = -1;
    }
  else
    {
      /* constant stack returned - guess with sub and kludge invalid vals */
      /* pinfot.p_stack     = procsinfo_dat.pi_ru.ru_isrss; */
      pinfot.p_stack     =
	pinfot.p_size[0] - pinfot.p_data - procsinfo_dat.pi_tsize;
      if( pinfot.p_stack < 0 )
	{
	  pinfot.p_stack = 0;
	}
      pinfot.p_majflt    = procsinfo_dat.pi_majflt;
      pinfot.p_minflt    = procsinfo_dat.pi_minflt;
      pinfot.p_size_lim  = procsinfo_dat.pi_rlimit[RLIMIT_AS].rlim_cur ==
	RLIM_INFINITY ? -1.0 : procsinfo_dat.pi_rlimit[RLIMIT_AS].rlim_cur;
      pinfot.p_rss_lim   = procsinfo_dat.pi_rlimit[RLIMIT_RSS].rlim_cur ==
	RLIM_INFINITY ? -1.0 : procsinfo_dat.pi_rlimit[RLIMIT_RSS].rlim_cur;
      pinfot.p_data_lim  = procsinfo_dat.pi_rlimit[RLIMIT_DATA].rlim_cur ==
	RLIM_INFINITY ? -1.0 : procsinfo_dat.pi_rlimit[RLIMIT_DATA].rlim_cur;
      pinfot.p_stack_lim = procsinfo_dat.pi_rlimit[RLIMIT_STACK].rlim_cur ==
	RLIM_INFINITY ? -1.0 : procsinfo_dat.pi_rlimit[RLIMIT_STACK].rlim_cur;
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size[0] == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      PROCMON_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_IRIX */
/* ---- */
/* IRIX */
/* ---- */
#if defined( _PROCMON_IRIX )
#include <sys/types.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h>
int PROCMON_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  long long num_chars; /* number of chars to print in strncpy */
  PROCMON_INFO_struct pinfot; /* if no error, will be coppied to pinfo */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  int proc_fid; /* ioctl file id */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct rminfo	rminfo_sysmp; /* sysmp */
  struct rusage ruse_getrusage; /* getrusage */
  prpsinfo_t
    prpsinfo_info; /* ioctl(fd, PIOCPSINFO, &prpsinfo_info) */
  prusage_t
    prusage_info; /* ioctl(fd, PIOCUSAGE,  &prusage_info) */
  /* init */
  PROCMON_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* sysmp */
  sysmp(MP_SAGET, MPSA_RMINFO, &rminfo_sysmp, sizeof( struct rminfo ));
  pinfot.m_size = rminfo_sysmp.physmem * pagesize;
  pinfot.m_free = rminfo_sysmp.freemem * pagesize;
  /* /proc/<PID> */
  sprintf( proc_buf_filename, "/proc/%llu", pinfot.p_pid );
  proc_fid = open( proc_buf_filename, O_RDONLY );
  if( proc_fid == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone/inaccessible */
      pinfot.p_ppid    = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
      pinfot.p_time[0] = 0;
      pinfot.wall      = 0;
      pinfot.p_wall[0] = 0;
      pinfot.p_majflt  = 0;
      pinfot.p_minflt  = 0;
    }
  else
    {
      ioctl( proc_fid, PIOCPSINFO, &prpsinfo_info );
      ioctl( proc_fid, PIOCUSAGE,  &prusage_info );
      close( proc_fid );
      pinfot.p_ppid     = prpsinfo_info.pr_ppid;
      pinfot.p_uid      = prpsinfo_info.pr_uid;
      num_chars = PROCMON_MAX_LEN-1;
      strncpy( pinfot.p_command, prpsinfo_info.pr_fname, num_chars );
      pinfot.p_size[0]  = prusage_info.pu_size * pagesize;
      pinfot.p_rss      = prusage_info.pu_rss * pagesize;
      pinfot.p_time[0]  = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_utime ) + 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_stime );
      pinfot.wall       = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_tstamp );
      pinfot.p_wall[0]  = pinfot.wall - 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_starttime );
      pinfot.p_majflt  = prusage_info.pu_majf;
      pinfot.p_minflt  = prusage_info.pu_minf;
    }
  /* ------------------------------------------------------------------ */
  /* if getting info about this process, can call getrusage to get info */
  /* ------------------------------------------------------------------ */
  if( pid == 0 )
    {
      /* I thought this would work...just returns 0 though... */
      getrusage( RUSAGE_SELF, &ruse_getrusage );
      pinfot.p_data  = ruse_getrusage.ru_idrss;
      pinfot.p_stack = ruse_getrusage.ru_isrss;
    }
  else
    {
      pinfot.p_data  = 0;
      pinfot.p_stack = 0;
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.num_samples_printed = pinfo->num_samples_printed;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size[0] == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      PROCMON_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_IRIX */
/* ----- */
/* ALPHA */
/* ----- */
#if defined( _PROCMON_ALPHA )
#include <sys/table.h>
#include <sys/user.h>
#include <sys/stat.h>
int PROCMON_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  long long num_chars; /* number of chars to print in strncpy */
  PROCMON_INFO_struct pinfot; /* if no error, will be coppied to pinfo */
  int lerr; /* local error return value */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  int proc_fid; /* ioctl file id */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct user user_tbl; /* table - TBL_UAREA */
  struct tbl_pmemstats pmemstats_tbl; /* table - TBL_PMEMSTATS */
  struct tbl_vmstats vmstats_tbl; /* table - TBL_VMSTATS */
  prpsinfo_t
    prpsinfo_info; /* ioctl(fd, PIOCPSINFO, &prpsinfo_info) */
  prusage_t
    prusage_info; /* ioctl(fd, PIOCUSAGE,  &prusage_info) */
  struct stat file_stat; /* stat on /proc/<file> */
  /* init */
  PROCMON_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* table: TBL_PMEMSTATS */
  table( TBL_PMEMSTATS, 0, &pmemstats_tbl, 1,
         sizeof( struct tbl_pmemstats ) );
  pinfot.m_size = pmemstats_tbl.physmem;
  /* table: TBL_VMSTATS: */
  /* http://h30097.www3.hp.com/wbem/docs/t64upds.html : FreePhysicalMemory */
  table( TBL_VMSTATS, 0, &vmstats_tbl, 1,
         sizeof( struct tbl_vmstats ) );
  pinfot.m_free = pinfot.m_size -
    pagesize * ( 
                vmstats_tbl.wire_count +
                vmstats_tbl.active_count +
                vmstats_tbl.inactive_count +
                0*vmstats_tbl.ubc_pages /* currently set to 0.
                                         * On some nodes, top reporting
                                         * numbers like 32M free.  My guess
                                         * is that the system is periodically
                                         * taking memory but not reporting
                                         * freeing it up.  In my parallel_leak
                                         * test, ignoring the ubc pages gave
                                         * the best results for standard
                                         * malloc leak.  With the ubc pages,
                                         * reported 32M free - without gave
                                         * reasonable numbers. Dunno.... */
                );
  /* /proc/<PID> */
  sprintf( proc_buf_filename, "/proc/%llu", pinfot.p_pid );
  proc_fid = open( proc_buf_filename, O_RDONLY );
  if( proc_fid == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone/inaccessible */
      pinfot.p_ppid    = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
      pinfot.p_time[0] = 0;
      pinfot.wall      = 0;
      pinfot.p_wall[0] = 0;
      pinfot.p_majflt  = 0;
      pinfot.p_minflt  = 0;
    }
  else
    {
      ioctl( proc_fid, PIOCPSINFO, &prpsinfo_info );
      ioctl( proc_fid, PIOCUSAGE,  &prusage_info );
      /* for xterm processes, these have incorrect uid - do fstat on file */
      /* pinfot.p_uid     = prpsinfo_info.pr_uid; */
      fstat( proc_fid, &file_stat );
      pinfot.p_uid = file_stat.st_uid;
      close( proc_fid );
      pinfot.p_ppid    = prpsinfo_info.pr_ppid;
      num_chars =
        FNAME_SZ < PROCMON_MAX_LEN - 1 ? FNAME_SZ : PROCMON_MAX_LEN - 1;
      strncpy( pinfot.p_command, prpsinfo_info.pr_fname, num_chars );
      pinfot.p_size[0] = prpsinfo_info.pr_size * pagesize;
      pinfot.p_rss     = prpsinfo_info.pr_rssize * pagesize;
      pinfot.p_time[0] = PROCMON_TIMEN_TO_DOUBLE( prpsinfo_info.pr_time );
      pinfot.wall      = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pr_tstamp );
      pinfot.p_wall[0] = pinfot.wall - 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pr_create );
      pinfot.p_majflt  = prusage_info.pr_majf;
      pinfot.p_minflt  = prusage_info.pr_minf;
    }
  /* table: TBL_UAREA */
  /* ---------------------------------------------------------------- */
  /* when accessing another pid (eg not this process), table seems to */
  /* return a -1.  So, do not use this.                               */
  /* ---------------------------------------------------------------- */
  lerr = table( TBL_UAREA, pinfot.p_pid, &user_tbl, 1, sizeof( struct user ) );
  if( lerr == -1 )
    {
      pinfot.p_data    = 0;
      pinfot.p_stack   = 0;
    }
  else
    {
      pinfot.p_data  = user_tbl.u_dsize * pagesize;
      /* on machines I've tested, the u_ssize is wrong - assign by subtr */
      /* pinfot.p_stack = user_tbl.u_ssize * pagesize; */
      pinfot.p_stack = pinfot.p_size[0] - pinfot.p_data - user_tbl.u_tsize;
      /* use ioclt call below for getting this - works for getting info */
      /* about another pid                                              */
      /* 
         pinfot.p_size  = user_tbl.u_tsize * pagesize +
         pinfot.p_data + pinfot.p_stack;
      */
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.num_samples_printed = pinfo->num_samples_printed;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size[0] == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      PROCMON_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_ALPHA */
/* ----- */
/* LINUX */
/* ----- */
#if defined( _PROCMON_LINUX )
#define JIFFIE 100.0
#define KBYTE 1024.0
#include <ctype.h>
static inline char* skip_token(
                               const char *p
                               )
{
  while (p && isspace(*p)) p++;
  while (p && !isspace(*p)) p++;
  while (p && isspace(*p)) p++;
  return (char *)p;
}
static inline char* skip_ws(
                            const char *p
                            )
{
  while (p && isspace(*p)) p++;
  return (char *)p;
}
static inline long long read_buf(
                                 FILE *filep,
                                 char **file_buf,
                                 long long *file_buf_size
                                 )
{
  long long i = 0;
  do
    {
      if( i >= *file_buf_size -1 )
        {
          *file_buf_size += getpagesize();
          *file_buf = realloc( *file_buf, *file_buf_size );
        }
    }
  while( ((*file_buf)[i++] = fgetc(filep)) != EOF );
  if( i > 0 )
    {
      (*file_buf)[i-1] = '\000';
    }
  return( i );
}
int PROCMON_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  PROCMON_INFO_struct pinfot; /* temp coppied if no error */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  char *proc_buf; /* /proc name and file contents */
  long long num_chars; /* number of chars to print in strncpy */
  long long proc_buf_size; /* size of buffer */
  char *p; /* pointer into proc_buf */
  char *p_tmp; /* temp pointer into proc_buf */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct timespec time_current; /* clock_gettime */
  FILE *proc_fp; /* file pointer */
  double uptime;
  /* init */
  proc_buf_size = 0;
  proc_buf = NULL;
  PROCMON_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  clock_gettime( CLOCK_REALTIME, &time_current );
  pinfot.wall = PROCMON_TIMEN_TO_DOUBLE( time_current );
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* ------------- */
  /* /proc/meminfo */
  /* ------------- */
  sprintf( proc_buf_filename, "%s/%s", "/proc", "meminfo" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* cannot read info */
      pinfot.m_size = 0;
      pinfot.m_free = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = p != NULL ? strchr( p, '\n' ) : NULL;       /* end of line */
      p = skip_token( p );                            /* Mem: */
      p = skip_ws( p );                               /* spaces */
      pinfot.m_size = p != NULL ? strtod( p, &p ) : 0;/* total -> m_size */ 
      p = skip_token( p );                            /* used */
      pinfot.m_free = p != NULL ? strtod( p, &p ) : 0;/* free -> m_free */
    }
  /* ------------------ */
  /* /proc/<pid>/status */
  /* ------------------ */
  sprintf( proc_buf_filename, "%s/%llu/%s", "/proc", pinfot.p_pid, "status" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* process is gone/inaccessible */
      pinfot.p_ppid    = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
      pinfot.p_data    = 0;
      pinfot.p_stack   = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = p != NULL ? strstr(p, "Name:" ) : NULL;                 /* Name */
      p = skip_token(p);
      p_tmp = p != NULL ? strchr(p, '\n') : NULL;
      num_chars = p_tmp - p;
      num_chars =
        num_chars < PROCMON_MAX_LEN - 1 ? num_chars : PROCMON_MAX_LEN;
      strncpy( pinfot.p_command, p, num_chars );
      p = p != NULL ? strchr(p, '\n') : NULL;         /* get passed Name */
      p = p != NULL ? strstr(p, "PPid:" ) : NULL;                 /* PPid */
      p = skip_token(p);
      pinfot.p_ppid    = p != NULL ? strtoul( p, &p, 10 ) : 0;
      p = p != NULL ? strstr(p, "Uid:" ) : NULL;                  /* Uid */
      p = skip_token(p);
      pinfot.p_uid    = p != NULL ? strtoul( p, &p, 10 ) : 0;
      p = p != NULL ? strstr(p, "VmSize:" ) : NULL;               /* VmSize */
      p = skip_token(p);
      pinfot.p_size[0] = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmRSS:" ) : NULL;                /* VmRSS */
      p = skip_token(p);
      pinfot.p_rss     = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmData:" ) : NULL;               /* VmData */
      p = skip_token(p);
      pinfot.p_data    = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmStk:" ) : NULL;                /* VmStk */
      p = skip_token(p);
      pinfot.p_stack   = p != NULL ? strtod( p, &p ) * KBYTE : 0;
    }
  /* ---------------- */
  /* /proc/uptime     */
  /* needed by status */
  /* ---------------- */
  sprintf( proc_buf_filename, "%s/%s", "/proc", "uptime" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* cannot read info */
      uptime = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      uptime = p != NULL ? strtod( p, &p ) : 0; /* time since reboot */
    }
  /* ------------------------------------------ */
  /* /proc/<pid>/stat                           */
  /* Most not used - status (above) used mostly */
  /* ------------------------------------------ */
  sprintf( proc_buf_filename, "%s/%llu/%s", "/proc", pinfot.p_pid, "stat" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* process is gone/inaccessible */
      pinfot.p_time[0] = 0;
      pinfot.p_wall[0] = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
      pinfot.p_majflt  = 0;
      pinfot.p_minflt  = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = skip_token(p);                               /* 1  pid */
      p = p != NULL ? strchr(p, ')') + 1 : NULL;       /* 2  command */
      p = skip_token(p);                               /* 3  state */
      p = skip_token(p);                               /* 4  ppid */
      p = skip_token(p);                               /* 5  pgrp */
      p = skip_token(p);                               /* 6  session */
      p = skip_token(p);                               /* 7  tty */
      p = skip_token(p);                               /* 8  tpgid */
      p = skip_token(p);                               /* 9  flags */
      pinfot.p_minflt  =
        p != NULL ? strtod( p, &p ) : 0;               /* 10 minflt */
      p = skip_token(p);                               /* 11 cminflt */
      pinfot.p_majflt  =
        p != NULL ? strtod( p, &p ) : 0;               /* 12 majflt */
      p = skip_token(p);                               /* 13 cmajflt */
      pinfot.p_time[0]  =
        p != NULL ? strtod( p, &p )/JIFFIE : 0;        /* 14 utime */
      pinfot.p_time[0] +=
        p != NULL ? strtod( p, &p )/JIFFIE : 0;        /* 15 stime */
      p = skip_token(p);                               /* 16 cutime */
      p = skip_token(p);                               /* 17 cstime */
      p = skip_token(p);                               /* 18 counter */
      p = skip_token(p);                               /* 19 priority */
      p = skip_token(p);                               /* 20 timeout */
      p = skip_token(p);                               /* 21 itrealvalue */
      pinfot.p_wall[0]  =
        p != NULL ? -strtod( p, &p )/JIFFIE : 0;       /* 22 starttime */
      pinfot.p_wall[0] += uptime;
      pinfot.p_size[0]  =
        p != NULL ? strtod( p, &p ) : 0;               /* 23 vsize -> p_size */
      pinfot.p_rss   =
        p != NULL ? strtod( p, &p ) * pagesize : 0;    /* 24 rss -> p_rss */
    }
  /* ------------------------------------------------------------------ */
  /* Temp fix just to get working on bproc.  Seems that values for wall */
  /* time is using /proc/uptime of front end and /proc/<pid>/stat wrt   */
  /* back end.  This then gives p_wall bad value (currently negative)   */
  /* So, if get negative value, just set p_wall to that of p_time.      */
  /* ------------------------------------------------------------------ */
  if( pinfot.p_wall[0] < 0 )
    {
      pinfot.p_wall[0] = pinfot.p_time[0];
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.num_samples_printed = pinfo->num_samples_printed;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size[0] == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      PROCMON_info_copy( &pinfot, pinfo );
    }
  if( proc_buf != NULL )
    {
      free( proc_buf );
      proc_buf = NULL;
    }
  return( ierr );
}
#endif /* _PROCMON_LINUX */
/* ========================================================================

   Name
   ====
   PROCMON_init_file

   Purpose
   =======
   Initializes structs and opens file.
   Grouped funtionality needed to be called if new process ID (eg a fork
   was called so a new output file needs to be used).

   Arguments
   =========
   pinfo         Intent: inout
                 C type: PROCMON_INFO_struct *
                 struct to init

   psettings     Intent: inout
                 C type: PROCMON_INFO_SETTINGS_struct *
                 struct to init

   fp            Intent: inout
                 C type: FILE **
                 file opened (non NULL/stdout files will be closed)

   Return Values
   =============
   Non-0 if error

   Program Flow
   ============
   1) close fp if needed
   2) init structs
   3) open new file
   4) print settings

   ======================================================================== */
int PROCMON_init_file(
                      PROCMON_INFO_struct *pinfo,
                      PROCMON_INFO_SETTINGS_struct *psettings,
                      FILE **fp
                      )
{
  int
    ierr = 0; /* error return value */
  /* --------------- */
  /* set some things */
  /* --------------- */
  if( *fp != NULL && *fp != stdout )
    {
      fclose( *fp );
      *fp = NULL;
    }
  PROCMON_info_init( pinfo );
  PROCMON_info_settings_get( 0, NULL, psettings );
  if( strlen(psettings->outfile_name) > 0 )
    {
      *fp = fopen( psettings->outfile_name, "a" );
    }
  else
    {
      *fp = stdout;
    }
  PROCMON_info_settings_print( psettings, *fp );
  return( ierr );
}
/* ---------------------------------------------------------------- */
/* The following routines are interfaces into the rld functionality */
/* These routines can only be used if you have linked in            */
/*    libprocmon_rld.so                                             */
/* ---------------------------------------------------------------- */
/* ========================================================================

   Name
   ====
   PROCMON_rld_handler

   Purpose
   =======
   Get/print process info info - This is the function that the signal
   handler calls.  It can be called manually as well.

   Arguments
   =========
   sig           Intent: in
                 C type: int
                 The signal sent to this process.
                 If called independently, the value should be set to 0

   info          Intent: in
                 C type: siginfo_t *
                 One of the required args for sa_handler
                 Set to NULL if calling manually.

   context       Intent: in
                 C type: void *
                 One of the required args for sa_handler
                 Set to NULL if calling manually.

   Return Values
   =============
   none

   Program Flow
   ============
   1) get info
   2) print info

   ======================================================================== */
void PROCMON_rld_handler_noop(
                              int sig,
                              siginfo_t *info,
                              void *context
                              )
{
}
void PROCMON_rld_handler(
                         int sig,
                         siginfo_t *info,
                         void *context
                         )
{
  int procmon_error=0; /* error return value for procmon calls */
  /* --------------------------------------- */
  /* return if in handler or not initialized */
  /* --------------------------------------- */
  if( procmon_s.in_handler != 0 || ! procmon_s.initialized )
    return;
  procmon_s.in_handler = 1;
  /* ------------------------------------------------------------ */
  /* set to ignore signal (do not want to get signals backing up) */
  /* ------------------------------------------------------------ */
  if( sig != 0 )
    {
      if( sigaction(procmon_s.which_sigaction,
                    &(procmon_s.action_ignore), NULL ) != 0 )
        abort();
    }
  /* --------------------------------------------------------- */
  /* fork workaround - re-init to get new file/timer if needed */
  /* --------------------------------------------------------- */
  PROCMON_rld_reset();
  /* -------------- */
  /* get pinfo data */
  /* -------------- */
  while( (procmon_error = PROCMON_info_get( 0, &(procmon_s.pinfo) ) ) ==
         PROCMON_WARNING ) {};
  /* ------------------------ */
  /* print info               */
  /* file exists and no error */
  /*  - correct interval (or) */
  /*  - final output          */
  /* ------------------------ */
  if( ( procmon_s.filename_fp != NULL ) &&
      ( procmon_error == 0 ) && 
      ( ( procmon_s.psettings.interval > 0 &&
          procmon_s.pinfo.num_samples % procmon_s.psettings.interval == 0 ) ||
        ( sig == 0 ) ) )
    {
      PROCMON_info_print( &(procmon_s.pinfo), &(procmon_s.psettings),
                          procmon_s.filename_fp );
    }
  /* --------------------------------------------- */
  /* chain to previous sig catcher if given signal */
  /* probably do not want to do this...but might   */
  /* --------------------------------------------- */
  if( sig != 0 && procmon_s.chain_signal != 0 )
    {
      if( procmon_s.action_old.sa_flags & SA_SIGINFO &&
          procmon_s.action_old.sa_sigaction != NULL )
        {
          procmon_s.action_old.sa_sigaction(sig, info, context);
        }
      else if( procmon_s.action_old.sa_handler != NULL )
        {
          procmon_s.action_old.sa_handler(sig);
        }
    }
  /* --------------------------------------------------- */
  /* DONE: chain to previous sig catcher if given signal */
  /* --------------------------------------------------- */
  /* ----------------------- */
  /* no longer ingore signal */
  /* ----------------------- */
  if( sig != 0 )
    {
      if (sigaction(procmon_s.which_sigaction,
                    &(procmon_s.action), NULL ) != 0 )
        abort();
    }
  procmon_s.in_handler = 0;
}
/* ========================================================================

   Name
   ====
   PROCMON_rld_init_timer

   Purpose
   =======
   Sets up the timer PROCMON will use to get samples.
   No effect if not using automatic sampling (via
   the run time linker environment variable or
   relinking the executable with libprocmon_rld.so.

   Arguments
   =========
   void          Intent: in
                 C type: void
                 no arguments or return values

   Return Values
   =============
   none

   Program Flow
   ============
   1) initialize signal timer for rld method

   ======================================================================== */
void PROCMON_rld_init_timer( void )
{
  if( procmon_s.initialized == 0 )
    return;
#if defined(_PROCMON_ITIMER)
  /* ------ */
  /* itimer */
  /* ------ */
  if( setitimer(procmon_s.which_itimer, &(procmon_s.value),
                &(procmon_s.value_old)) != 0 )
    abort();
  /* ------------ */
  /* POSIX4 timer */
  /* ------------ */
#else
  if( timer_create( CLOCK_REALTIME, &(procmon_s.psxsigevent),
                &(procmon_s.psxtimer) ) != 0 )
    abort();
  if( timer_settime( procmon_s.psxtimer, 0, &(procmon_s.psxspec),
                 &(procmon_s.psxspec_old) ) != 0 )
    abort();
#endif
}
/* ========================================================================

   Name
   ====
   PROCMON_rld_reset

   Purpose
   =======
   This resets the automatic sampling environment
   after a fork() is called.  Some MPI implementations
   use fork() to spawn processes.  Various things
   need to be reset (eg a new output file and a
   new signal timer).  If you do not call this
   after the fork() (eg after the MPI_Init() call) when
   doing automatic sampling, the child process(es) will
   not do any automatic sampling until the last
   sampling at program termination (note, any manual
   sampling will still be done).

   Arguments
   =========
   void          Intent: in
                 C type: n/a
                 no calling arguments

   Return Values
   =============
   none

   Program Flow
   ============
   1) PROCMON_init_file
   2) PROCMON_rld_init_timer

   ======================================================================== */
void PROCMON_rld_reset()
{
  if( procmon_s.initialized && getpid() != procmon_s.psettings.pid )
    {
      PROCMON_init_file( &(procmon_s.pinfo), &(procmon_s.psettings),
                         &(procmon_s.filename_fp) );
      PROCMON_rld_init_timer();
    }
}
void procmon_rld_reset(
                       void
                       )
{
  PROCMON_rld_reset();
}
void procmon_rld_reset_(
                        void
                        )
{
  PROCMON_rld_reset();
}
void procmon_rld_reset__(
                         void
                         )
{
  PROCMON_rld_reset();
}
void PROCMON_RLD_RESET(
                       void
                       )
{
  PROCMON_rld_reset();
}
